#include "DAQDataFormats/RawMilliSliceIndex.h"
#include "DAQDataFormats/DataFormatException.h"

namespace daqdataformats{

//------------------------------------------------------------
// Default constructor
RawMilliSliceIndex::RawMilliSliceIndex(const uint32_t size):
     RawDAQData(0, (const init_t) &RawMilliSliceIndex::init)
   ,_Header(NULL)
{
    _function_sizeofdata = (sizeofdata_t) &RawMilliSliceIndex::sizeofdata;
    reserveInternalBuffer(size);
    defaultConstructorFloatingSize();
}

//------------------------------------------------------------------------------
RawMilliSliceIndex::RawMilliSliceIndex(const RawMilliSliceIndex& copy_in):
     RawDAQData(copy_in)
   ,_Header(NULL)
{
    lastOperationsInDeepCopy(copy_in);
}

//------------------------------------------------------------------------------
RawMilliSliceIndex::RawMilliSliceIndex(const void* millisliceheader_buffer):
     RawDAQData(0, (const init_t) &RawMilliSliceIndex::init)
   ,_Header(NULL)
{
    _function_sizeofdata = (sizeofdata_t) &RawMilliSliceIndex::sizeofdata;
   
    copy(millisliceheader_buffer);
}
//------------------------------------------------------------
// Destructor
RawMilliSliceIndex::~RawMilliSliceIndex() {
    if(_Header) delete _Header;
}

//------------------------------------------------------------
// Initialize the data members
void RawMilliSliceIndex::init(){

    /// After using resetBufferSource header is already constructed
    /// Now we need to initialize it
    _Header->init();
}

//------------------------------------------------------------------------------
// Getters
uint32_t RawMilliSliceIndex::getTime        (const uint32_t iindex) const {
        const uint32_t word = iindex * getSizeOfOneIndex() + rawmilliindx::WORD_Time;
        return BITFIELDGET(RAWBUFF32[word], rawmilliindx::MASK_Time, rawmilliindx::SHIFT_Time);
}

//------------------------------------------------------------------------------
uint32_t RawMilliSliceIndex::getOffset     (const uint32_t iindex) const {
        const uint32_t word = iindex * getSizeOfOneIndex() + rawmilliindx::WORD_Offset;
        return BITFIELDGET(RAWBUFF32[word], rawmilliindx::MASK_Offset, rawmilliindx::SHIFT_Offset);
}
//------------------------------------------------------------------------------
uint32_t RawMilliSliceIndex::getIndexNumber() const { return _Header->getIndexNumber();}

//------------------------------------------------------------------------------
bool RawMilliSliceIndex::isClosed() const {return _Header->isClosed();}

//------------------------------------------------------------------------------
// Add an index
bool RawMilliSliceIndex::addIndex  (const uint32_t time_length, const uint32_t offset){

    if(_shouldBufferBeInternal){
        addToInternalBuffer(getSizeOfOneIndex());
        /// Check pointers after each read or resize buffer
        checkBufferInternalness();
    }

    const uint32_t index = getIndexNumber();
    setTime(index, time_length);
    setOffset(index, offset);
    return advanceIndexNumber();
}

//------------------------------------------------------------------------------
// Setters
// Set time of the "iindex"-th index
bool RawMilliSliceIndex::setTime     (const uint32_t iindex, const uint32_t time_length){
        const uint32_t word = iindex * getSizeOfOneIndex() + rawmilliindx::WORD_Time;
        BITFIELDSET(RAWBUFF32[word], time_length, rawmilliindx::MASK_Time, rawmilliindx::SHIFT_Time);
        return true;
}

//------------------------------------------------------------------------------
// Set Offset of the "iindex"-th index
bool RawMilliSliceIndex::setOffset   (const uint32_t iindex, const uint32_t offset){
        const uint32_t word = iindex * getSizeOfOneIndex() + rawmilliindx::WORD_Offset;
        BITFIELDSET(RAWBUFF32[word], offset, rawmilliindx::MASK_Offset, rawmilliindx::SHIFT_Offset);
        return true;
}

//------------------------------------------------------------------------------
bool RawMilliSliceIndex::setIndexNumber(uint32_t indx){return _Header->setIndexNumber(indx);}

//------------------------------------------------------------------------------
bool RawMilliSliceIndex::advanceIndexNumber(){ return _Header->advanceIndexNumber();}

//------------------------------------------------------------------------------
bool RawMilliSliceIndex::setClosed(bool closeness){return _Header->setClosed(closeness);}

//------------------------------------------------------------------------------
bool RawMilliSliceIndex::setBufferSource(const void* pointer){
    _Buffer = (void*) pointer;

    _shouldBufferBeInternal = isInternalBuffer();

   /// If Header is already defined, then set its buffer source
   /// Otherwise, need to construct it
   if(_Header)return _Header->setBufferSource(pointer);

   _Header = new RawMilliSliceIndexHeader(pointer); ///< version of the NanoSliceHeader equals the version of the NanoSlice

   return true;
}

    //------------------------------------------------------------------------------
// Print the MilliSlice header information
void RawMilliSliceIndex::print(std::ostream& os) const{

    _Header->print(os);

    os<<"\nMilliSliceIndex:";

    const uint32_t size = sizeofdata();

    for(uint32_t i = 0; i < size; ++i){
        os<<" ";
        printWord(i,false, os);
    }

    os<<"\nMilliSliceIndex RAW bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size; ++i){
        os<<std::endl;
        printWord(i,true, os);
    }
}

//------------------------------------------------------------------------------
// Size of _Buffer in words
uint32_t RawMilliSliceIndex::sizeofdata() const {

    /// If header is currently not defined, just return 1. Eventually it will be defined
    if(!_Header) return 1;

    uint32_t size = _Header->sizeofdata()
                  + getIndexNumber() * getSizeOfOneIndex();

    if(isClosed()) size += _Header->sizeofdata();

    return size;
}

//------------------------------------------------------------------------------
// Close MilliSlice Index
bool RawMilliSliceIndex::close(){

    if(isClosed())return false;

    /// Set Close bit in the header
    setClosed(true);

    /// Copy Header into the end of the MilliSlice Index
    readDataGeneral(_Header->getBuffer(), _Header->sizeofdata());

    /// Check pointers after each read or resize buffer
    checkBufferInternalness();

    return true;
}


//------------------------------------------------------------------------------
// This is a general method to read the microslice from millislice_buffer to _Buffer
void* RawMilliSliceIndex::readData(const void* buffer)
{
    void* buffer_copy = (void*) buffer;

    const uint32_t millisliceindex_header_size = _Header->sizeofdata();

    /// Read the header and find out the size of the millislice
    buffer_copy = readDataGeneral(buffer_copy, millisliceindex_header_size, 0);

    /// Check pointers after each read
    /// To be able to use sizeofdata()
    checkBufferInternalness();

    /// Now, read the rest of the millislice
    buffer_copy = readDataGeneral(buffer_copy, sizeofdata() - millisliceindex_header_size, millisliceindex_header_size);

    /// Check pointers after each read or resize buffer
    checkBufferInternalness();

    return buffer_copy;
 }// end of readData(void*)

}// end of namespace
