#include <DatabaseUtils/DAQConfig/PixelEnableMasks.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <algorithm>

namespace dbutils {
namespace daqconfig {

PixelEnableMasks::
PixelEnableMasks(const SubsystemNameSpec& nameSpec,
                 std::vector<std::string> dcmList)
{
    std::vector<int64_t> subsystemConfigIdList;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList);
    if (! status) {return;}
    std::sort(dcmList.begin(), dcmList.end());

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       dcmList[0], dcmList[dcmList.size()-1]);
        //std::cout << dcmList[0] << std::endl;

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("mask");

        std::string fname("NovaDatabase/tables/DAQConfig/FEBEnableMasks.xml");
        _febEnableMaskDataSet.
            reset(new ConfigDataTree(fname, subsystemConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       dcmList[0], dcmList[dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("mask");

        std::string fname("NovaDatabase/tables/DAQConfig/PixelEnableMasks.xml");
        _pixelEnableDataSet.
            reset(new ConfigDataTree(fname, subsystemConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    _hasValidData = _febEnableMaskDataSet.get() != 0 &&
        _febEnableMaskDataSet->hasValidData() &&
        _pixelEnableDataSet.get() != 0 &&
        _pixelEnableDataSet->hasValidData();
}

PixelEnableMasks::
PixelEnableMasks(const int64_t& globalConfigId,
                 std::vector<std::string> dcmList)
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    std::vector<int64_t> subsystemConfigIdList;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 DCM_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 HARDWARE_CONFIG_TYPE,
                                                 subsystemConfigIdList);
    if (! status) {return;}
    std::sort(dcmList.begin(), dcmList.end());

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       dcmList[0], dcmList[dcmList.size()-1]);
        //std::cout << dcmList[0] << std::endl;

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("mask");

        std::string fname("NovaDatabase/tables/DAQConfig/FEBEnableMasks.xml");
        _febEnableMaskDataSet.
            reset(new ConfigDataTree(fname, subsystemConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       dcmList[0], dcmList[dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("mask");

        std::string fname("NovaDatabase/tables/DAQConfig/PixelEnableMasks.xml");
        _pixelEnableDataSet.
            reset(new ConfigDataTree(fname, subsystemConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    _hasValidData = _febEnableMaskDataSet.get() != 0 &&
        _febEnableMaskDataSet->hasValidData() &&
        _pixelEnableDataSet.get() != 0 &&
        _pixelEnableDataSet->hasValidData();
}

int PixelEnableMasks::
getEnabledPixelCountForDCM(const std::string& dcmName)
{
    int count = 0;
    for (int ifeb = 0; ifeb < 64; ++ifeb) {
        count += getEnabledPixelCountForFEB(dcmName, ifeb);
    }
    return count;
}

bool PixelEnableMasks::
getFEBEnableMask(const std::string& dcmName, int64_t& febMask)
{
    if (_febEnableMaskDataSet.get() == 0) {return false;}
    return _febEnableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                           febMask);
}

bool PixelEnableMasks::
febIsEnabled(const std::string& dcmName, const int& febId)
{
    if (_febEnableMaskDataSet.get() == 0) {return false;}
    int64_t febMask;
    if (! _febEnableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                          febMask)) {
        return false;
    }

    int64_t _64BitOne = 1;
    return ((febMask & (_64BitOne << febId)) != 0);
}

int PixelEnableMasks::
getEnabledPixelCountForFEB(const std::string& dcmName,
                           const int& febId)
{
    if (! this->febIsEnabled(dcmName, febId)) {return 0;}

    if (_pixelEnableDataSet.get() == 0) {return 0;}
    int32_t pixelMask;
    if (! _pixelEnableDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                         "mask", pixelMask)) {
        return 0;
    }

    int count = 0;
    for (int ipix = 0; ipix < 32; ++ipix) {
        if ((pixelMask & (1 << ipix)) != 0) {
            ++count;
        }
    }
    return count;
}

bool PixelEnableMasks::
getPixelEnableMask(const std::string& dcmName,
                   const int& febId, int32_t& pixelMask)
{
    if (_pixelEnableDataSet.get() == 0) {return false;}
    return _pixelEnableDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                          "mask", pixelMask);
}

bool PixelEnableMasks::
pixelIsEnabled(const std::string& dcmName,const int& febId,const int& pixelId)
{
    if (_pixelEnableDataSet.get() == 0) {return false;}
    int32_t pixelMask;
    if (! _pixelEnableDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                         "mask", pixelMask)) {
        return false;
    }

    return ((pixelMask & (1 << pixelId)) != 0);
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
