#include <NovaTimingUtilities/TimingUtilities.h>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <iostream>
#include <string>

namespace NDTU = novadaq::timeutils;
namespace BPT = boost::posix_time;

/**
 * Command-line utility for converting between various time representations.
 */

int main(int argc, char **argv) {
    int inputOption = 0;
    uint64_t inputNovaTime = 0;
    struct timespec inputTimespec;
    inputTimespec.tv_sec = 0;
    inputTimespec.tv_nsec = 0;
    std::string inputTimeString = "";
    std::string partialTimeString = "";
    uint64_t inputPSec = 0;
    bool inputUTC = false;

    // parse command line argument(s).
    bool printUsage = false;
    if (argc < 2) {
        printUsage = true;
    }
    else {
        try {
            std::string argString(argv[1]);
            if (argString.find(":") != std::string::npos) {
                inputOption = 1;
                inputTimeString = argString;

                std::string::size_type pos = argString.find("UTC");
                if (pos != std::string::npos) {
                    argString = argString.substr(0, pos);
                    inputUTC = true;
                }
                boost::trim(argString);

                pos = argString.find(".");
                partialTimeString = argString.substr(0, pos);

                if (pos != std::string::npos) {
                    std::string fracString = argString.substr(pos+1);
                    fracString.append("000000000000");
                    fracString = fracString.substr(0, 12);
                    inputPSec =
                        boost::lexical_cast<uint64_t>(fracString);
                }
            }
            else if (argString.find(".") != std::string::npos) {
                inputOption = 2;
                int pos = argString.find(".");
                inputTimespec.tv_sec =
                    boost::lexical_cast<time_t>(argString.substr(0, pos));

                std::string fracString = argString.substr(pos+1);
                fracString.append("000000000");
                fracString = fracString.substr(0, 9);
                inputTimespec.tv_nsec =
                    boost::lexical_cast<long>(fracString);
            }
            else {
                inputOption = 3;
                if (argString.find("x") != std::string::npos) {
                    std::istringstream iss(argString);
                    iss >> std::hex >> inputNovaTime;
                }
                else {
                    inputNovaTime = boost::lexical_cast<uint64_t>(argString);
                }
            }
        }
        catch (boost::bad_lexical_cast& excpt) {
            printUsage = true;
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cerr << "Usage: " << argv[0] << " <inputTime> " << std::endl;
        std::cerr << "  where <inputTime> is one of the following:"
                  << std::endl;
        std::cerr << "   - a human readable datetime string that is converted "
                  << "to a UNIX timespec and a NOvA base time" << std::endl;
        std::cerr << "     * this should have a format of \"yyyy-mmm-dd "
                  << "HH:MM:SS.frac[ UTC]\" [quotes needed]" << std::endl;
        std::cerr << "     * the fractional part of seconds supports "
                  << "values down to one picosecond, .000000000001"
                  << std::endl;
        std::cerr << "     * if UTC is not specified, then the time is "
                  << "interpreted as LOCAL time" << std::endl;
        std::cerr << "   - UNIX timespec components that are converted "
                  << "to a human readable string and a NOvA base time"
                  << std::endl;
        std::cerr << "     * this should have a format of \"sec.frac\" "
                  << "[quotes not needed, but the dot is needed]"
                  << std::endl;
        std::cerr << "     * the fractional part of seconds supports "
                  << "values down to one nanosecond, .000000001"
                  << std::endl;
        std::cerr << "   - a NOvA base time value that is converted "
                  << "to a human readable string and a UNIX timespec"
                  << std::endl;
        std::cerr << "     * this should be a simple number with no "
                  << "punctuation (hex format supported)" << std::endl;
        std::cerr << "  examples:"
                  << std::endl;
        std::cerr << "    " << argv[0] << " \"2010-Jul-22 9:01\""
                  << " (datetime string)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " \"2010-Jul-22 14:22:30."
                  << "000000015625\""
                  << " (datetime string)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " \"2010-Jul-22 00:00 UTC\""
                  << " (datetime string)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " 1120208640000000"
                  << " (NOvA time)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " 0x43d950a5b0000"
                  << " (NOvA time)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " 1279807260.005000"
                  << " (UNIX time, seconds.nanoseconds)"
                  << std::endl;
        std::cerr << "    " << argv[0] << " 1279807260.0"
                  << " (UNIX time, seconds.nanoseconds)"
                  << std::endl;
        return -1;
    }

    // handle the various possible times
    std::cout << std::endl;

    if (inputOption == 3) {
        struct timespec outputTimespec;
        NDTU::convertNovaTimeToUnixTime(inputNovaTime, outputTimespec);
        std::string outputDateString =
            NDTU::convertNovaTimeToString(inputNovaTime);
        std::cout << "A NOvA base time of " << inputNovaTime
                  << " corresponds to..." << std::endl;
        std::cout << "  a UNIX time of " << outputTimespec.tv_sec
                  << " sec, " << outputTimespec.tv_nsec << " nsec"
                  << std::endl;
        std::cout << "  a calendar date of " << outputDateString
                  << std::endl;
    }

    if (inputOption == 2) {
        uint64_t outputNovaTime;
        bool convertStatus =
            NDTU::convertUnixTimeToNovaTime(inputTimespec, outputNovaTime);
        std::string outputDateString =
            NDTU::convertUnixTimeToString(inputTimespec);
        std::cout << "A UNIX time of " << inputTimespec.tv_sec
                  << " sec, " << inputTimespec.tv_nsec << " nsec"
                  << " corresponds to..." << std::endl;
        if (convertStatus) {
            std::cout << "  a NOvA base time of " << outputNovaTime
                      << std::endl;
        }
        else {
            std::cout << "  an invalid NOvA base time" << std::endl;
        }
        std::cout << "  a calendar date of " << outputDateString
                  << std::endl;
    }

    if (inputOption == 1) {
        BPT::ptime posixTime = BPT::time_from_string(partialTimeString);
        struct tm blah = BPT::to_tm(posixTime);
        time_t unix_seconds;
        if (inputUTC) {
            unix_seconds = timegm(&blah);
        }
        else {
            unix_seconds = mktime(&blah);
        }
        std::string timeString = "";

        // brute-force way of getting UTC offset
        //BPT::ptime ptime1(BPT::second_clock::local_time());
        //BPT::ptime ptime2(BPT::second_clock::universal_time());
        //BPT::time_duration gmtDiff = ptime1 - ptime2;
        //unix_seconds += gmtDiff.total_seconds();

        std::cout << "A time string of " << inputTimeString
                  << " corresponds to..." << std::endl;
        if (unix_seconds >= NDTU::NOVA_EPOCH) {
	    struct timespec tt={unix_seconds,inputPSec/1000};
	    uint64_t novaTime;
	    NDTU::convertUnixTimeToNovaTime( tt, novaTime );
# if 0
            double val = NDTU::NOVA_TIME_FACTOR *
                (((double) unix_seconds - NDTU::NOVA_EPOCH) +
                 (((double) inputPSec) / 1E12));
            uint64_t novaTime = (uint64_t) val;
# endif
            std::cout << "  a NOvA base time of " << novaTime
                      << std::endl;
            timeString = NDTU::convertNovaTimeToString(novaTime);
        }
        else {
            std::cout << "  an invalid NOvA base time" << std::endl;
        }

        std::cout << "  a UNIX time of " << unix_seconds
                  << " sec, " << ((uint32_t) (inputPSec / 1000)) << " nsec"
                  << std::endl;

        if (timeString.length() == 0) {
            struct timespec unixTime;
            unixTime.tv_sec = unix_seconds;
            unixTime.tv_nsec = ((long) (inputPSec / 1000));
            timeString = NDTU::convertUnixTimeToString(unixTime);
        }
        std::cout << "  a calendar date of " << timeString
                  << std::endl;
    }

    std::cout << std::endl;
    return 0;
}
